#include <ESP8266WiFi.h>
#include <ESP8266WebServer.h>
#include <Wire.h>
#include <Adafruit_MPU6050.h>
#include <Adafruit_Sensor.h>

// WiFi credentials
const char* ssid = "Wifi Name";  
const char* password = "Password";
+

// Flex sensor setup
const int flexPin = A0;                // NodeMCU analog pin A0
const float VCC = 3.3;                 // NodeMCU voltage (3.3V)
const float R_DIV = 47000.0;           // 47K ohm resistor
const float flatResistance = 25000.0;  // Resistance when flat
const float bendResistance = 100000.0; // Resistance at 90 deg

// Click detection variables
const float CLICK_MIN_ANGLE = 60.0;    // Minimum angle to trigger a click
const float CLICK_MAX_ANGLE = 90.0;    // Maximum angle to consider
const unsigned long CLICK_COOLDOWN = 700; // Milliseconds to wait before registering another click
boolean isClicked = false;             // Current click state
boolean clickSent = false;             // Whether a click was already sent for this bend
unsigned long lastClickTime = 0;       // When the last click occurred

ESP8266WebServer server(80);
Adafruit_MPU6050 mpu;

// Debug flag
bool debug = true;

void setup() {
  // Initialize serial communication
  Serial.begin(115200);
  delay(100);  // Short delay for serial to initialize
  
  Serial.println("\n\n=== Gesture Mouse Control System ===");
  Serial.println("Initializing...");
  
  // Initialize I2C for MPU6050
  Wire.begin(D2, D1);  // SDA=D2, SCL=D1
  delay(50);  // Give some time for I2C to initialize

  // Initialize MPU6050
  Serial.println("Connecting to MPU6050...");
  bool mpuInitialized = false;
  for (int i = 0; i < 5; i++) {  // Try 5 times to initialize
    if (mpu.begin()) {
      mpuInitialized = true;
      break;
    }
    Serial.println("Failed to find MPU6050 chip. Retrying...");
    delay(500);
  }
  
  if (!mpuInitialized) {
    Serial.println("ERROR: Could not connect to MPU6050 sensor!");
    Serial.println("Please check your wiring and restart the device.");
    while (1) {
      delay(10);
    }
  }
  
  Serial.println("MPU6050 Found and Initialized Successfully!");
  
  // Configure MPU6050 sensor settings
  mpu.setAccelerometerRange(MPU6050_RANGE_8_G);
  mpu.setGyroRange(MPU6050_RANGE_500_DEG);
  mpu.setFilterBandwidth(MPU6050_BAND_21_HZ);
  
  // Test reading from flex sensor
  Serial.println("Testing flex sensor...");
  int flexReading = analogRead(flexPin);
  Serial.print("Flex sensor raw reading: ");
  Serial.println(flexReading);
  if (flexReading == 0 || flexReading > 1020) {
    Serial.println("WARNING: Flex sensor reading seems invalid. Check your connections!");
  } else {
    Serial.println("Flex sensor reading looks good!");
  }
  
  // Connect to WiFi
  WiFi.mode(WIFI_STA);  // Set WiFi to station mode
  WiFi.disconnect();    // Disconnect from any previous connections
  delay(100);
  
  Serial.print("Connecting to WiFi network: ");
  Serial.println(ssid);
  
  WiFi.begin(ssid, password);
  
  // Wait for connection with timeout
  int timeout = 0;
  while (WiFi.status() != WL_CONNECTED && timeout < 20) {
    delay(500);
    Serial.print(".");
    timeout++;
  }
  
  if (WiFi.status() != WL_CONNECTED) {
    Serial.println("\nFAILED to connect to WiFi network!");
    Serial.println("Please check your WiFi credentials and signal strength.");
    Serial.println("The system will continue to function, but web server will not be available.");
  } else {
    Serial.println("");
    Serial.print("Connected successfully to WiFi network: ");
    Serial.println(ssid);
    Serial.print("IP address: ");
    Serial.println(WiFi.localIP());
    
    // Define server routes
    server.on("/", HTTP_GET, handleRoot);
    server.on("/data", HTTP_GET, handleData);
    server.onNotFound(handleNotFound);
    
    // Start server
    server.begin();
    Serial.println("HTTP server started on port 80");
    Serial.println("You can access the data endpoint at: http://" + WiFi.localIP().toString() + "/data");
  }
  
  Serial.println("System initialization complete. Starting main loop...");
}

void loop() {
  // Handle client requests if WiFi is connected
  if (WiFi.status() == WL_CONNECTED) {
    server.handleClient();
  } else {
    // If WiFi disconnected, try to reconnect occasionally
    static unsigned long lastReconnectAttempt = 0;
    unsigned long currentMillis = millis();
    
    if (currentMillis - lastReconnectAttempt > 30000) { // Try every 30 seconds
      lastReconnectAttempt = currentMillis;
      
      if (WiFi.status() != WL_CONNECTED) {
        Serial.println("WiFi disconnected. Attempting to reconnect...");
        WiFi.reconnect();
      }
    }
  }
  
  // Always update the flex sensor state regardless of WiFi status
  updateFlexSensorState();
  
  // Small delay to prevent CPU hogging
  delay(10);
}

// Custom map function for float values
float mapf(float x, float in_min, float in_max, float out_min, float out_max) {
  return (x - in_min) * (out_max - out_min) / (in_max - in_min) + out_min;
}

void updateFlexSensorState() {
  // Read the flex sensor
  int ADCflex = analogRead(flexPin);
  
  // Calculate voltage and resistance
  float Vflex = ADCflex * VCC / 1023.0;
  float Rflex = R_DIV * (VCC / Vflex - 1.0);
  
  // Calculate the bend angle
  float angle = mapf(Rflex, flatResistance, bendResistance, 0, 90.0);
  
  // Constrain angle to avoid out-of-range values
  angle = constrain(angle, 0, 90.0);
  
  // Get current time
  unsigned long currentTime = millis();
  
  // Debug output (uncomment if needed)
  static unsigned long lastDebugOutput = 0;
  if (debug && (currentTime - lastDebugOutput > 1000)) {
    lastDebugOutput = currentTime;
    Serial.print("Flex Sensor - Raw: ");
    Serial.print(ADCflex);
    Serial.print(", Angle: ");
    Serial.print(angle);
    Serial.print(" degrees, Click state: ");
    Serial.println(isClicked ? "TRUE" : "false");
  }
  
  // Check if angle is in the click range
  if (angle >= CLICK_MIN_ANGLE && angle <= CLICK_MAX_ANGLE) {
    // If not already clicked and enough time has passed since last click
    if (!clickSent && (currentTime - lastClickTime > CLICK_COOLDOWN)) {
      isClicked = true;
      clickSent = true;
      lastClickTime = currentTime;
      Serial.println("CLICK EVENT DETECTED! Angle: " + String(angle));
    }
  } else {
    // Reset click sent flag when no longer in click range
    clickSent = false;
    isClicked = false;
  }
}

void handleRoot() {
  String html = "<!DOCTYPE html>\n";
  html += "<html>\n";
  html += "<head>\n";
  html += "<title>Gesture Mouse Control</title>\n";
  html += "<meta http-equiv='refresh' content='5'>\n"; // Auto-refresh page every 5 seconds
  html += "<style>\n";
  html += "body { font-family: Arial, sans-serif; margin: 20px; background-color: #f0f0f0; }\n";
  html += "h1 { color: #0066cc; }\n";
  html += ".status { padding: 10px; background-color: #e0e0e0; border-radius: 5px; margin-top: 20px; }\n";
  html += ".data { font-family: monospace; margin-top: 20px; }\n";
  html += "</style>\n";
  html += "</head>\n";
  html += "<body>\n";
  html += "<h1>Gesture Mouse Control System</h1>\n";
  
  html += "<div class='status'>\n";
  html += "<p>System Status: <strong>Running</strong></p>\n";
  html += "<p>Server IP: <strong>" + WiFi.localIP().toString() + "</strong></p>\n";
  html += "<p>Get sensor data at: <a href='/data'>/data</a> (JSON format)</p>\n";
  html += "</div>\n";

  // Get current sensor data
  sensors_event_t a, g, temp;
  mpu.getEvent(&a, &g, &temp);
  
  int ADCflex = analogRead(flexPin);
  float Vflex = ADCflex * VCC / 1023.0;
  float Rflex = R_DIV * (VCC / Vflex - 1.0);
  float angle = mapf(Rflex, flatResistance, bendResistance, 0, 90.0);
  angle = constrain(angle, 0, 90.0);
  
  html += "<div class='data'>\n";
  html += "<h2>Current Sensor Data:</h2>\n";
  
  html += "<p>Accelerometer (m/s²):<br>\n";
  html += "X: " + String(a.acceleration.x, 2) + "<br>\n";
  html += "Y: " + String(a.acceleration.y, 2) + "<br>\n";
  html += "Z: " + String(a.acceleration.z, 2) + "</p>\n";
  
  html += "<p>Gyroscope (rad/s):<br>\n";
  html += "X: " + String(g.gyro.x, 2) + "<br>\n";
  html += "Y: " + String(g.gyro.y, 2) + "<br>\n";
  html += "Z: " + String(g.gyro.z, 2) + "</p>\n";
  
  html += "<p>Flex Sensor:<br>\n";
  html += "Raw: " + String(ADCflex) + "<br>\n";
  html += "Angle: " + String(angle, 2) + " degrees<br>\n";
  html += "Click state: " + String(isClicked ? "TRUE" : "false") + "</p>\n";
  
  html += "<p>Temperature: " + String(temp.temperature, 2) + " °C</p>\n";
  html += "</div>\n";
  
  html += "</body>\n";
  html += "</html>\n";
  
  server.send(200, "text/html", html);
}

void handleData() {
  sensors_event_t a, g, temp;
  mpu.getEvent(&a, &g, &temp);
  
  // Read the flex sensor for the JSON data
  int ADCflex = analogRead(flexPin);
  float Vflex = ADCflex * VCC / 1023.0;
  float Rflex = R_DIV * (VCC / Vflex - 1.0);
  float angle = mapf(Rflex, flatResistance, bendResistance, 0, 90.0);
  angle = constrain(angle, 0, 90.0);
  
  // Enable CORS headers
  server.sendHeader("Access-Control-Allow-Origin", "*");
  server.sendHeader("Access-Control-Allow-Methods", "GET");
  server.sendHeader("Access-Control-Allow-Headers", "Content-Type");
  
  // Create JSON response
  String json = "{";
  json += "\"accelerometer\": {";
  json += "\"x\": " + String(a.acceleration.x) + ", ";
  json += "\"y\": " + String(a.acceleration.y) + ", ";
  json += "\"z\": " + String(a.acceleration.z);
  json += "}, ";
  json += "\"gyroscope\": {";
  json += "\"x\": " + String(g.gyro.x) + ", ";
  json += "\"y\": " + String(g.gyro.y) + ", ";
  json += "\"z\": " + String(g.gyro.z);
  json += "}, ";
  json += "\"flex\": {";
  json += "\"angle\": " + String(angle) + ", ";
  json += "\"raw\": " + String(ADCflex) + ", ";
  json += "\"click\": " + String(isClicked ? "true" : "false");
  json += "}, ";
  json += "\"temperature\": " + String(temp.temperature);
  json += "}";
  
  server.send(200, "application/json", json);
  
  // Debug output
  if (debug) {
    Serial.println("Data endpoint accessed - Sent JSON response");
  }
}

void handleNotFound() {
  String message = "File Not Found\n\n";
  message += "URI: ";
  message += server.uri();
  message += "\nMethod: ";
  message += (server.method() == HTTP_GET) ? "GET" : "POST";
  message += "\nArguments: ";
  message += server.args();
  message += "\n";
  
  for (uint8_t i = 0; i < server.args(); i++) {
    message += " " + server.argName(i) + ": " + server.arg(i) + "\n";
  }
  
  server.send(404, "text/plain", message);
}